# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import random
from typing import Any

from torch.optim import Adam, AdamW

from libinephany.observations import observation_utils
from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.observations.observers.base_observers import GlobalObserver
from libinephany.observations.observers.global_observers.constants import LHOPT_CONSTANTS
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates
from libinephany.utils.enums import ModelFamilies

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class InitialHyperparameters(GlobalObserver):

    def __init__(self, pad_with: float = 0.0, **kwargs) -> None:
        """
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(**kwargs)

        self.pad_with = pad_with

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """

        if self.include_hparams is None:
            raise ValueError(f"{self.__class__.__name__} must be provided with include_hparams.")

        available_hparams = HyperparameterStates.get_all_hyperparameters()

        return len([hparam for hparam in available_hparams if hparam in self.include_hparams])

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    @property
    def can_inform(self) -> bool:
        """
        :return: Whether observations from the observer can be used in the agent info dictionary.
        """

        return False

    @property
    def requires_include_hparams(self) -> bool:
        """
        :return: Whether the observation requires include_hparams to be provided.
        """

        return True

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.VECTOR

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        assert self.include_hparams is not None

        initial_internal_values = hyperparameter_states.get_initial_internal_values(self.include_hparams)
        self._cached_observation = initial_internal_values
        initial_internal_values_list = [
            self.pad_with if initial_internal_value is None else initial_internal_value
            for hparam_name, initial_internal_value in initial_internal_values.items()
            if hparam_name in self.include_hparams
        ]
        return initial_internal_values_list

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class OptimizerTypeOneHot(GlobalObserver):

    OPTIMS = [Adam.__name__, AdamW.__name__]

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """

        return len(self.OPTIMS)

    @property
    def can_inform(self) -> bool:
        """
        :return: Whether observations from the observer can be used in the agent info dictionary.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.VECTOR

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        optimizer_type = self.observer_config.optimizer_name

        if optimizer_type not in self.OPTIMS:
            index = None

        else:
            index = self.OPTIMS.index(optimizer_type)

        return observation_utils.create_one_hot_observation(vector_length=self.vector_length, one_hot_index=index)

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class ModelFamilyOneHot(GlobalObserver):

    UNIT_EPISODE = "episode"
    UNIT_TIMESTEP = "timestep"

    def __init__(
        self,
        *,
        zero_vector_chance: float = 0.2,
        zero_vector_frequency_unit: str = "episode",
        **kwargs,
    ) -> None:
        """
        :param zero_vector_chance: Chance of the output vector being masked with zeros.
        :param zero_vector_frequency_unit: Unit of time to sample the zero vector.
        :param kwargs: Miscellaneous keyword arguments.
        """
        super().__init__(**kwargs)
        self.should_zero = False

        assert 0.0 <= zero_vector_chance < 1.0
        self.zero_vector_chance = zero_vector_chance
        self._sample_zero_vector()

        if zero_vector_frequency_unit not in [self.UNIT_EPISODE, self.UNIT_TIMESTEP]:
            raise ValueError(f"Unknown zero_vector_frequency_unit: {zero_vector_frequency_unit}")

        self.zero_vector_frequency_unit = zero_vector_frequency_unit
        self.family_vector = self._create_family_vector()

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """

        return len(ModelFamilies)

    @property
    def can_inform(self) -> bool:
        """
        :return: Whether observations from the observer can be used in the agent info dictionary.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.VECTOR

    def _create_family_vector(self) -> list[float]:
        """
        :return: Creates and returns the model family one-hot vector.
        """

        family_name = self.observer_config.nn_family_name
        known_name = family_name in (family.value for family in ModelFamilies)

        if known_name:
            family_idx = ModelFamilies.get_index(family_name)

        else:
            family_idx = None

        return observation_utils.create_one_hot_observation(vector_length=self.vector_length, one_hot_index=family_idx)

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        if not self.in_training_mode:
            return self.family_vector

        if self.zero_vector_frequency_unit == self.UNIT_TIMESTEP:
            self._sample_zero_vector()

        if self.should_zero:
            return [0.0 for _ in range(self.vector_length)]

        else:
            return self.family_vector

    def _sample_zero_vector(self) -> None:
        """
        Determines whether the output vector of this observer should be masked with zeros.
        """
        self.should_zero = random.choices([True, False], [self.zero_vector_chance, (1 - self.zero_vector_chance)])[0]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}

    def reset(self) -> None:
        """
        Resets the observer.
        """

        self._sample_zero_vector()


class LHOPTHyperparameterRatio(GlobalObserver):
    """
    LHOPT-specific hyperparameter ratio observer that returns the ratio of current value to initial value
    for all hyperparameter actions.

    This observer computes: current_value / initial_value for each hyperparameter,
    providing insights into how much hyperparameters have changed from their starting values.
    """

    def __init__(self, pad_with: float = 0.0, **kwargs) -> None:
        """
        :param include_hparams: Names of the hyperparameters to include in the initial values vector returned by
        this observation.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(**kwargs)

        self.pad_with = pad_with

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """

        if self.include_hparams is None:
            raise ValueError(f"{self.__class__.__name__} must be provided with include_hparams.")

        available_hparams = HyperparameterStates.get_all_hyperparameters()

        return len([hparam for hparam in available_hparams if hparam in self.include_hparams])

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    @property
    def can_inform(self) -> bool:
        """
        :return: Whether observations from the observer can be used in the agent info dictionary.
        """

        return False

    @property
    def requires_include_hparams(self) -> bool:
        """
        :return: Whether the observation requires include_hparams to be provided.
        """

        return True

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.VECTOR

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        assert self.include_hparams is not None

        # Get initial and current hyperparameter values
        initial_values = hyperparameter_states.get_initial_internal_values(self.include_hparams)
        initial_values = {
            hparam_name: self.pad_with if initial_value is None else initial_value
            for hparam_name, initial_value in initial_values.items()
            if hparam_name in self.include_hparams
        }
        current_values = hyperparameter_states.get_current_internal_values(self.include_hparams)
        current_values = {
            hparam_name: self.pad_with if current_value is None else current_value
            for hparam_name, current_value in current_values.items()
            if hparam_name in self.include_hparams
        }

        ratios = []

        for hparam_name in initial_values.keys():
            initial_value = initial_values[hparam_name]
            current_value = current_values[hparam_name]

            if initial_value is None or current_value is None:
                ratios.append(0.0)
                continue

            if abs(initial_value) < LHOPT_CONSTANTS["ZERO_DIVISION_TOLERANCE"]:
                ratios.append(0.0)
            else:
                ratios.append(current_value / initial_value)

        return ratios

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from abc import abstractmethod
from typing import Any

import numpy as np
from loguru import logger

from libinephany.utils import random_seeds
from libinephany.utils.transforms import (
    HYPERPARAMETER_TRANSFORM_REGISTRY,
    HyperparameterTransform,
    HyperparameterTransformType,
)

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class Sampler:

    def __init__(
        self,
        lower_bound: float | int | None,
        upper_bound: float | int | None,
        sample_dtype: type[np.generic | float | int | str] = np.float64,
        **kwargs,
    ) -> None:
        """
        :param lower_bound: Lower bound of the distribution to sample from.
        :param upper_bound: Upper bound of the distribution to sample from.
        :param sample_dtype: Datatype the returned samples should have.
        :param kwargs: Miscellaneous keyword arguments.
        """

        if upper_bound is not None and lower_bound is not None and upper_bound < lower_bound:
            upper_bound, lower_bound = lower_bound, upper_bound

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.sample_dtype = sample_dtype

    def __call__(
        self,
        number_of_samples: int = 1,
        seed: int | None = None,
    ) -> list[Any] | np.ndarray | float:
        """
        :param number_of_samples: Number of samples to make.
        :param seed: Random seed to use for distribution sampling.
        :return: List/Array of sampled values or a single sampled value.
        """

        if seed is not None:
            random_seeds.set_all_seeds(seed)

        sample = self.sample(number_of_samples=number_of_samples)

        if isinstance(sample, np.ndarray) and sample.ndim > 1:
            return sample.squeeze(0)

        elif isinstance(sample, np.ndarray) and sample.size == 1:
            return sample.item()

        elif isinstance(sample, list) and len(sample) == 1:
            return sample[0]

        return sample

    @abstractmethod
    def _sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray | list[Any]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        raise NotImplementedError

    def sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray | list[Any]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        samples = self._sample(number_of_samples=number_of_samples, **kwargs)

        if self.lower_bound is not None or self.upper_bound is not None:
            samples = np.clip(np.array(samples), a_min=self.lower_bound, a_max=self.upper_bound)
            samples = samples.astype(self.sample_dtype)

        return samples

    @classmethod
    def get_subclasses(cls):
        """Recursively gets subclasses of the Sampler class."""
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass


class LogUniformSampler(Sampler):

    def _sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray | list[Any]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return sample_from_log_uniform(
            lower_bound=self.lower_bound, upper_bound=self.upper_bound, number_of_samples=number_of_samples
        )


class LogNormalSampler(Sampler):

    def __init__(
        self,
        lower_bound: float | int | None,
        upper_bound: float | int | None,
        mean: float,
        sigma: float,
        **kwargs,
    ) -> None:
        """
        :param step: Difference between each value in the discrete range.
        :param lower_bound: Lower bound of the distribution to sample from.
        :param upper_bound: Upper bound of the distribution to sample from.
        :param mean: Mean of the underlying normal distribution.
        :param sigma: Standard deviation of the underlying normal distribution.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(lower_bound=lower_bound, upper_bound=upper_bound, **kwargs)

        self.mean = mean
        self.sigma = sigma

    def _sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray | list[Any]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return sample_from_log_normal(
            mean=self.mean,
            sigma=self.sigma,
            lower_bound=self.lower_bound,
            upper_bound=self.upper_bound,
            number_of_samples=number_of_samples,
        ).astype(self.sample_dtype)


class TransformedNormalSampler(Sampler):

    def __init__(
        self,
        lower_bound: float | int | None,
        upper_bound: float | int | None,
        transform: str,
        mean: float,
        sigma: float,
        **kwargs,
    ) -> None:
        """
        :param step: Difference between each value in the discrete range.
        :param lower_bound: Lower bound of the distribution to sample from.
        :param upper_bound: Upper bound of the distribution to sample from.
        :param transform: Name of the hyperparameter transform to apply.
        :param mean: Mean of the underlying normal distribution.
        :param sigma: Standard deviation of the underlying normal distribution.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(lower_bound=lower_bound, upper_bound=upper_bound, **kwargs)

        self.transform = HYPERPARAMETER_TRANSFORM_REGISTRY[HyperparameterTransformType(transform)]()
        self.mean = mean
        self.sigma = sigma

    def _sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray | list[Any]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return sample_from_transformed_normal(
            transform=self.transform,
            mean=self.mean,
            sigma=self.sigma,
            lower_bound=self.lower_bound,
            upper_bound=self.upper_bound,
            number_of_samples=number_of_samples,
        ).astype(self.sample_dtype)


class UniformSampler(Sampler):

    def _sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray | list[Any]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return sample_from_uniform(
            lower_bound=self.lower_bound, upper_bound=self.upper_bound, number_of_samples=number_of_samples
        )


class DiscreteRangeSampler(Sampler):

    def __init__(
        self,
        step: float,
        lower_bound: float | int,
        upper_bound: float | int,
        **kwargs,
    ) -> None:
        """
        :param step: Difference between each value in the discrete range.
        :param lower_bound: Lower bound of the distribution to sample from.
        :param upper_bound: Upper bound of the distribution to sample from.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(lower_bound=lower_bound, upper_bound=upper_bound, **kwargs)

        self.step = step

    def _sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray | list[Any]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return sample_from_discrete_range(
            lower_bound=self.lower_bound,
            upper_bound=self.upper_bound,
            number_of_samples=number_of_samples,
            step=self.step,
        ).astype(self.sample_dtype)


class RoundRobinSampler(Sampler):

    def __init__(self, lower_bound: float | int, upper_bound: float | int, **kwargs) -> None:
        """
        :param lower_bound: Lower bound of the distribution to sample from.
        :param upper_bound: Upper bound of the distribution to sample from.
        :param kwargs: Miscellaneous keyword arguments.
        """
        super().__init__(lower_bound=lower_bound, upper_bound=upper_bound, **kwargs)

        self.magnitude_list = self.generate_magnitude_list()
        self.sampled_elements: set[float] = set()

    def generate_magnitude_list(self) -> list[float]:
        """
        :return: List of possible values that can be sampled in the sampler.
        """

        start_magnitude = math.floor(math.log10(self.upper_bound))
        end_magnitude = math.floor(math.log10(self.lower_bound))

        magnitude_list = []

        for magnitude in range(start_magnitude, end_magnitude - 1, -1):
            for i in range(1, 10):
                value = i * 10**magnitude

                if self.lower_bound <= value <= self.upper_bound:
                    magnitude_list.append(value)

        return magnitude_list

    def _sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray | list[Any]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        if len(self.sampled_elements) == len(self.magnitude_list):
            self.sampled_elements = set()

        sample_choices = [element for element in self.magnitude_list if element not in self.sampled_elements]
        sample = []

        for _ in range(number_of_samples):
            choice = np.random.choice(sample_choices, size=1).item()
            sample_choices.remove(choice)
            sample.append(choice)

        self.sampled_elements.update(sample)

        return np.array(sample)


class DiscreteValueSampler(Sampler):

    def __init__(
        self,
        discrete_values: list[float | int | str],
        sample_dtype: type[np.generic | float | int | str] = np.float64,
        lower_bound: float | int | None = None,
        upper_bound: float | int | None = None,
        **kwargs,
    ) -> None:
        """
        :param lower_bound: Lower bound of the distribution to sample from.
        :param upper_bound: Upper bound of the distribution to sample from.
        :param discrete_values: List of discrete values to sample from.
        :param sample_dtype: Datatype the returned samples should have.
        :param kwargs: Miscellaneous keyword arguments.
        """

        if sample_dtype == str or any(isinstance(value, str) for value in discrete_values):
            lower_bound, upper_bound = None, None

        else:
            lower_bound = min(discrete_values)  # type: ignore
            upper_bound = max(discrete_values)  # type: ignore

        super().__init__(lower_bound=lower_bound, upper_bound=upper_bound, sample_dtype=sample_dtype, **kwargs)  # type: ignore

        self.discrete_values = discrete_values

    def _sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray | list[Any]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return sample_from_discrete_values(
            discrete_values=self.discrete_values,
            number_of_samples=number_of_samples,
        ).astype(self.sample_dtype)


class DiscreteValueListSampler(DiscreteValueSampler):

    def __init__(
        self,
        length: int,
        discrete_values: list[float | int | str],
        sample_dtype: type[np.generic | float | int | str] = np.float64,
        **kwargs,
    ) -> None:
        """
        :param length: Length of list to sample.
        :param discrete_values: List of discrete values to sample from.
        :param sample_dtype: Datatype the returned samples should have.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(discrete_values=discrete_values, sample_dtype=sample_dtype, **kwargs)
        self.list_length = length

    def _sample(self, number_of_samples: int = 1, **kwargs) -> list[np.ndarray | list[Any]]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return [super()._sample(number_of_samples=self.list_length) for _ in range(number_of_samples)]


class RoundRobinDiscreteValueSampler(Sampler):

    def __init__(
        self,
        discrete_values: list[float | int],
        lower_bound: float | int | None = None,
        upper_bound: float | int | None = None,
        **kwargs,
    ) -> None:
        """
        :param discrete_values: List of discrete values to sample from.
        :param kwargs: Miscellaneous keyword arguments.
        """

        if lower_bound is not None or upper_bound is not None:
            logger.warning(
                f"{self.__class__.__name__} has been given bounds which are not None. This class "
                f"overrides these values and so they should not be manually set."
            )

        super().__init__(lower_bound=min(discrete_values), upper_bound=max(discrete_values), **kwargs)

        self.discrete_values = discrete_values

        self.sampled_elements: set[float] = set()

    def _sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray | list[Any]:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        if len(self.sampled_elements) == len(self.discrete_values):
            self.sampled_elements = set()

        sample_choices = [element for element in self.discrete_values if element not in self.sampled_elements]
        sample = []

        for _ in range(number_of_samples):
            choice = np.random.choice(sample_choices, size=1).item()
            sample_choices.remove(choice)
            sample.append(choice)

        self.sampled_elements.update(sample)

        return np.array(sample)


# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def build_sampler(sampler_name: str, lower_bound: float | int, upper_bound: float | int, **kwargs) -> Sampler:
    """
    :param sampler_name: Name of the sampler to build.
    :param lower_bound: Smallest value that can be sampled by the sampler.
    :param upper_bound: Largest value that can be sampled by the sampler.
    :param kwargs: Miscellaneous keyword arguments that may be required depending on the sampler chosen.
    :return: Constructed sampler.
    """

    possible_samplers = {sampler_type.__name__: sampler_type for sampler_type in Sampler.get_subclasses()}

    try:
        return possible_samplers[sampler_name](lower_bound=lower_bound, upper_bound=upper_bound, **kwargs)  # type: ignore

    except KeyError as e:
        raise ValueError(
            f"Unrecognised sampler {sampler_name}! Possible values: {list(possible_samplers.keys())}."
        ) from e


def sample_from_log_uniform(
    lower_bound: float | int, upper_bound: float | int, number_of_samples: int = 1
) -> np.ndarray:
    """
    :param lower_bound: Lower bound of the distribution.
    :param upper_bound: Upper bound of the distribution.
    :param number_of_samples: Number of samples to make.
    :return: Sampled result.
    """

    sample = np.random.uniform(math.log(lower_bound), math.log(upper_bound), size=number_of_samples)
    sample = np.exp(sample)

    return sample


def sample_from_log_normal(
    mean: float | int,
    sigma: float | int,
    lower_bound: float | int | None,
    upper_bound: float | int | None,
    number_of_samples: int = 1,
) -> np.ndarray:
    """
    :param mean: Mean of the underlying normal distribution.
    :param sigma: Standard deviation of the underlying normal distribution.
    :param lower_bound: Lower bound of the distribution.
    :param upper_bound: Upper bound of the distribution.
    :param number_of_samples: Number of samples to make.
    :return: Sampled result.
    """

    sample = np.random.lognormal(mean=mean, sigma=sigma, size=number_of_samples)

    if (lower_bound is not None) or (upper_bound is not None):
        sample = np.clip(sample, a_min=lower_bound, a_max=upper_bound)

    return sample


def sample_from_transformed_normal(
    transform: HyperparameterTransform,
    mean: float | int,
    sigma: float | int,
    lower_bound: float | int | None,
    upper_bound: float | int | None,
    number_of_samples: int = 1,
) -> np.ndarray:
    """
    :param transform: HyperparameterTransform to apply to the samples.
    :param mean: Mean of the underlying normal distribution (internal value).
    :param sigma: Standard deviation of the underlying normal distribution (internal value).
    :param lower_bound: Lower bound of the distribution (external value).
    :param upper_bound: Upper bound of the distribution (external value).
    :param number_of_samples: Number of samples to make.
    :return: Sampled result.
    """

    sample = np.random.normal(mean, sigma, size=number_of_samples)
    sample = np.array(list(map(transform.to_external, sample)))
    if (lower_bound is not None) or (upper_bound is not None):
        sample = np.clip(sample, a_min=lower_bound, a_max=upper_bound)

    return sample


def sample_from_uniform(lower_bound: float | int, upper_bound: float | int, number_of_samples: int = 1) -> np.ndarray:
    """
    :param lower_bound: Lower bound of the distribution.
    :param upper_bound: Upper bound of the distribution.
    :param number_of_samples: Number of samples to make.
    :return: Sampled result.
    """

    return np.random.uniform(lower_bound, upper_bound, size=number_of_samples)


def sample_from_discrete_range(
    lower_bound: float | int, upper_bound: float | int, step: float, number_of_samples: int = 1
) -> np.ndarray:
    """
    :param lower_bound: Lower bound of the distribution.
    :param upper_bound: Upper bound of the distribution.
    :param step: Difference between each value in the discrete range.
    :param number_of_samples: Number of samples to make.
    :return: Sampled result.
    """

    sample = np.arange(lower_bound, upper_bound + step, step, dtype=np.float64)

    sample = np.random.choice(sample, size=number_of_samples)

    return sample


def sample_from_discrete_values(discrete_values: list[float | int | str], number_of_samples: int = 1) -> np.ndarray:
    """
    :param discrete_values: List of discrete values to sample from.
    :param number_of_samples: Number of samples to make.
    :return: Sampled result.
    """

    sample = np.random.choice(discrete_values, size=number_of_samples)

    return sample



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from abc import ABC, abstractmethod
from types import ModuleType
from typing import Any, final

from libinephany.observations.observers import global_observers, local_observers
from libinephany.observations.observers.base_observers import GlobalObserver, LocalObserver
from libinephany.pydantic_models.configs.observer_config import AgentObserverConfig, ObserverConfig
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates
from libinephany.utils.standardizers import Standardizer
from libinephany.utils.typing import ObservationInformation

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class ObserverContainer(ABC):

    def __init__(
        self,
        global_config: ObserverConfig,
        agent_config: AgentObserverConfig,
        standardizer: Standardizer | None,
        **kwargs,
    ) -> None:
        """
        :param global_config: ObserverConfig that can be used to inform various observation calculations.
        :param agent_config: AgentObserverConfig storing configuration of the agent's actions and observation space for
        this type of agent.
        :param standardizer: None or the standardizer to apply to the returned observations.
        :param kwargs: Miscellaneous kwargs.
        """

        self.global_config = global_config
        self.agent_config = agent_config
        self.standardizer = standardizer

        self._observers: list[LocalObserver | GlobalObserver] = self._build_observers()

    @property
    @abstractmethod
    def observer_config(self) -> dict[str, dict[str, Any] | None] | None:
        """
        :return: Config of the various observers to construct.
        """

        raise NotImplementedError

    @property
    @abstractmethod
    def observer_module(self) -> ModuleType:
        """
        :return: Module in which the observer classes are stored.
        """

        raise NotImplementedError

    @property
    @abstractmethod
    def required_observer_kwargs(self) -> dict[str, Any]:
        """
        :return: Dictionary of kwargs required by the observers in the set observer module.
        """

        raise NotImplementedError

    @final
    @property
    def total_observer_size(self) -> int:
        """
        :return: Total size of the observation vector this container produces.
        """

        return sum(observer.observation_size for observer in self._observers)

    @final
    def _build_observers(
        self,
    ) -> list[LocalObserver | GlobalObserver]:
        """
        :return: List of instantiated observers.
        """

        if self.observer_config is None:
            return []

        observers = []

        for observer_name, additional_observer_kwargs in self.observer_config.items():
            try:
                observer_type: type[LocalObserver | GlobalObserver] = getattr(self.observer_module, observer_name)

            except AttributeError as e:
                raise AttributeError(f"The class {observer_name} does not exist within {self.observer_module}!") from e

            if additional_observer_kwargs is None:
                additional_observer_kwargs = {}

            observer_kwargs = self.required_observer_kwargs
            observer_kwargs.update(**additional_observer_kwargs)

            observer = observer_type(
                standardizer=self.standardizer,
                observer_config=self.global_config,
                **observer_kwargs,
            )

            observers.append(observer)

        return observers

    @final
    def get_required_trackers(self) -> list[dict[str, dict[str, Any] | None]]:
        """
        :return: List of trackers required by each of the stored observers.
        """

        return [observer.get_required_trackers() for observer in self._observers]

    @final
    def observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        return_dict: bool = False,
        num_categories: int | None = None,
    ) -> tuple[list[float | int], dict[str, list[float | int] | dict[str, float | int]] | None]:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param return_dict: Whether to return a dictionary of observations as well as the normal vector.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Tuple of a list of floats or integers to add to the agent's observation vector and a dictionary of
        observations as well as the normal vector.
        """

        observations = []
        observations_dict: dict[str, list[float | int] | dict[str, float | int]] = {}

        for observer in self._observers:
            observer_obs, observer_obs_dict = observer.observe(
                observation_inputs=observation_inputs,
                hyperparameter_states=hyperparameter_states,
                tracked_statistics=tracked_statistics,
                action_taken=action_taken,
                return_dict=return_dict,
                num_categories=num_categories,
            )

            if return_dict and observer_obs_dict is not None:
                observations_dict[observer.__class__.__name__] = observer_obs_dict

            elif return_dict:
                observations_dict[observer.__class__.__name__] = observer_obs

            observations += observer_obs

        return observations, observations_dict if return_dict else None

    @final
    def inform(self) -> ObservationInformation:
        """
        :return: Dictionary mapping inform keys to observations to add to agent info.
        """

        informed = {}

        for observer in self._observers:
            observer_info = observer.inform()

            if observer_info is not None:
                informed[observer.__class__.__name__] = observer_info

        return informed

    @final
    def reset(self) -> None:
        """
        Resets all observers in the container.
        """

        for observer in self._observers:
            observer.reset()

    @final
    def train(self) -> None:
        """
        Sets all observers into training mode.
        """

        for observer in self._observers:
            observer.train()

    @final
    def infer(self) -> None:
        """
        Sets all observers into inference mode.
        """

        for observer in self._observers:
            observer.infer()


class LocalObserverContainer(ObserverContainer):

    def __init__(self, agent_id: str, parameter_group_name: str | None, **kwargs) -> None:
        """
        :param agent_id: ID of the agent this container serves.
        :param parameter_group_name: Name of the parameter group the agent that this container serves is controlling.
        :param kwargs: Kwargs for the base class.
        """

        self.agent_id = agent_id
        self.parameter_group_name = parameter_group_name

        super().__init__(**kwargs)

    @property
    def requires_validation_loss(self) -> bool:
        """
        :return: Whether the observation requires validation loss to be calculated.
        """

        return any(observer.requires_validation_loss for observer in self._observers)

    @property
    def observer_config(self) -> dict[str, dict[str, Any] | None] | None:
        """
        :return: Config of the various observers to construct.
        """

        return self.agent_config.local_observers

    @property
    def observer_module(self) -> ModuleType:
        """
        :return: Module in which the observer classes are stored.
        """

        return local_observers

    @property
    def required_observer_kwargs(self) -> dict[str, Any]:
        """
        :return: Dictionary of kwargs required by the observers in the set observer module.
        """

        return dict(
            agent_id=self.agent_id,
            parameter_group_name=self.parameter_group_name,
            number_of_discrete_actions=self.agent_config.number_of_discrete_actions,
            action_scheme_index=self.agent_config.action_scheme_index,
            number_of_action_schemes=self.agent_config.number_of_action_schemes,
        )


class GlobalObserverContainer(ObserverContainer):

    @property
    def observer_config(self) -> dict[str, dict[str, Any] | None] | None:
        """
        :return: Config of the various observers to construct.
        """

        return self.agent_config.global_observers

    @property
    def observer_module(self) -> ModuleType:
        """
        :return: Module in which the observer classes are stored.
        """

        return global_observers

    @property
    def required_observer_kwargs(self) -> dict[str, Any]:
        """
        :return: Dictionary of kwargs required by the observers in the set observer module.
        """

        return {}



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from abc import ABC, abstractmethod
from typing import Any

from libinephany.observations.observation_utils import StatisticStorageTypes, compute_cdf_feature
from libinephany.observations.observers.base_observers import GlobalObserver
from libinephany.observations.observers.global_observers.constants import LHOPT_CONSTANTS
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class LHOPTBaseObserver(GlobalObserver, ABC):
    """
    Base class for LHOPT outer step observers to eliminate duplicate code.
    """

    def __init__(
        self,
        decay_factor: float = LHOPT_CONSTANTS["DEFAULT_DECAY_FACTOR"],
        time_window: int = LHOPT_CONSTANTS["DEFAULT_TIME_WINDOW"],
        **kwargs,
    ) -> None:
        """
        :param decay_factor: Decay factor for CDF calculation in [1, 2.5, 5, 10, 20]
        :param time_window: Number of time steps to consider for CDF calculation
        :param kwargs: Other observation keyword arguments.
        """
        super().__init__(**kwargs)

        # Store time series data for CDF calculation
        self._time_series: list[tuple[float, float]] = []  # (time, value) pairs
        self._current_time: float = 0.0

        self.decay_factor = max(0.0, decay_factor)
        self.time_window = max(1, time_window)

    @property
    def can_standardize(self) -> bool:
        """
        This observer has its own CDF calculation, no need to standardize.
        :return: Whether the observation can be standardized.
        """
        return False

    @staticmethod
    def _compute_log_ratio(numerator: float, denominator: float) -> float:
        """
        Compute the log ratio.

        :param numerator: Numerator value
        :param denominator: Denominator value
        :return: Log ratio value
        """
        # Calculate the ratio of numerator to denominator
        invalid_denominator = math.isinf(denominator) or math.isnan(denominator)

        if denominator <= LHOPT_CONSTANTS["ZERO_DIVISION_TOLERANCE"] or invalid_denominator:
            return 0.0

        ratio = numerator / denominator

        if ratio <= 0:
            return 0.0

        return math.log(ratio)

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the StatisticStorageTypes
        enumeration class.
        """
        return StatisticStorageTypes.VECTOR

    def _compute_cdf_feature(self, value: float) -> float:
        """
        Compute CDF feature for the given value.
        training loss will be added to the time series after this call.
        :param value: The value to compute CDF feature for
        :return: CDF feature value
        """
        return compute_cdf_feature(value, self._time_series, self.decay_factor, self._current_time, self.time_window)

    def _update_time(self) -> None:
        """Update the current time counter."""
        self._current_time += 1.0

    @abstractmethod
    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        """

        ...

    def reset(self) -> None:
        """Reset the observer by clearing the time series."""
        self._time_series = []
        self._current_time = 0.0


class LHOPTCheckpointBaseObserver(GlobalObserver, ABC):
    """
    Base class for checkpoint-based observers to eliminate duplicate code.
    """

    def __init__(self, checkpoint_interval: int = LHOPT_CONSTANTS["DEFAULT_CHECKPOINT_INTERVAL"], **kwargs) -> None:
        """
        :param checkpoint_interval: How often to create checkpoints (in outer model steps).
        :param kwargs: Miscellaneous keyword arguments.
        """
        super().__init__(**kwargs)

        self._history: list[float] = []

        self.checkpoint_interval = checkpoint_interval
        self.last_value: float | None = None

    @property
    def can_standardize(self) -> bool:
        """
        This observer has its own CDF calculation, no need to standardize.
        :return: Whether the observation can be standardized.
        """
        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in.
        """
        return StatisticStorageTypes.FLOAT

    def _update_history(self, value: float) -> None:
        """
        Update the history with a new value and maintain sliding window.

        :param value: The new value to add to history
        """
        self._history.append(value)

        # Keep only the last checkpoint_interval values for sliding window
        if len(self._history) > self.checkpoint_interval:
            self._history = self._history[-self.checkpoint_interval :]

    def _should_create_checkpoint(self) -> bool:
        """
        Check if we should create a checkpoint.

        :return: True if checkpoint should be created, False otherwise
        """
        return len(self._history) >= self.checkpoint_interval

    def _cold_start(self, value: float) -> None:
        """
        Handle cold start by setting the last value if not already set.

        :param value: The value to set as last value if cold start
        """
        if self.last_value is None:
            self.last_value = value

    @abstractmethod
    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        """

        ...

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """
        return {}

    def reset(self) -> None:
        """Reset the observer by clearing history."""
        self._history = []
        self.last_value = None
